@@ -1,36 +1,98 @@
 mod state;
 
-pub use self::state::{MemoryStackState, MemoryStackSubstate, StackState};
+pub use self::state::{MemoryStackAccount, MemoryStackState, MemoryStackSubstate, StackState};
 
-use crate::gasometer::{self, Gasometer};
+use crate::gasometer::{self, Gasometer, StorageTarget};
 use crate::{
 	Capture, Config, Context, CreateScheme, ExitError, ExitReason, ExitSucceed, Handler, Opcode,
 	Runtime, Stack, Transfer,
 };
-use alloc::{rc::Rc, vec::Vec};
+use alloc::{
+	collections::{BTreeMap, BTreeSet},
+	rc::Rc,
+	vec::Vec,
+};
 use core::{cmp::min, convert::Infallible};
 use ethereum::Log;
 use primitive_types::{H160, H256, U256};
 use sha3::{Digest, Keccak256};
 
+macro_rules! emit_exit {
+	($reason:expr) => {{
+		let reason = $reason;
+		event!(Exit {
+			reason: &reason,
+			return_value: &Vec::new(),
+		});
+		reason
+	}};
+	($reason:expr, $return_value:expr) => {{
+		let reason = $reason;
+		let return_value = $return_value;
+		event!(Exit {
+			reason: &reason,
+			return_value: &return_value,
+		});
+		(reason, return_value)
+	}};
+}
+
 pub enum StackExitKind {
 	Succeeded,
 	Reverted,
 	Failed,
 }
 
+#[derive(Default, Clone, Debug)]
+struct Accessed {
+	accessed_addresses: BTreeSet<H160>,
+	accessed_storage: BTreeSet<(H160, H256)>,
+}
+
+impl Accessed {
+	pub fn access_address(&mut self, address: H160) {
+		self.accessed_addresses.insert(address);
+	}
+
+	pub fn access_addresses<I>(&mut self, addresses: I)
+	where
+		I: Iterator<Item = H160>,
+	{
+		for address in addresses {
+			self.accessed_addresses.insert(address);
+		}
+	}
+
+	pub fn access_storages<I>(&mut self, storages: I)
+	where
+		I: Iterator<Item = (H160, H256)>,
+	{
+		for storage in storages {
+			self.accessed_storage.insert((storage.0, storage.1));
+		}
+	}
+}
+
+#[derive(Clone, Debug)]
 pub struct StackSubstateMetadata<'config> {
 	gasometer: Gasometer<'config>,
 	is_static: bool,
 	depth: Option<usize>,
+	accessed: Option<Accessed>,
 }
 
 impl<'config> StackSubstateMetadata<'config> {
 	pub fn new(gas_limit: u64, config: &'config Config) -> Self {
+		let accessed = if config.increase_state_access_gas {
+			Some(Accessed::default())
+		} else {
+			None
+		};
 		Self {
 			gasometer: Gasometer::new(gas_limit, config),
 			is_static: false,
 			depth: None,
+			accessed,
 		}
 	}
 
@@ -39,6 +101,17 @@ impl<'config> StackSubstateMetadata<'config> {
 		self.gasometer
 			.record_refund(other.gasometer.refunded_gas())?;
 
+		if let (Some(mut other_accessed), Some(self_accessed)) =
+			(other.accessed, self.accessed.as_mut())
+		{
+			self_accessed
+				.accessed_addresses
+				.append(&mut other_accessed.accessed_addresses);
+			self_accessed
+				.accessed_storage
+				.append(&mut other_accessed.accessed_storage);
+		}
+
 		Ok(())
 	}
 
@@ -60,6 +133,7 @@ impl<'config> StackSubstateMetadata<'config> {
 				None => Some(0),
 				Some(n) => Some(n + 1),
 			},
+			accessed: self.accessed.as_ref().map(|_| Accessed::default()),
 		}
 	}
 
@@ -78,6 +152,36 @@ impl<'config> StackSubstateMetadata<'config> {
 	pub fn depth(&self) -> Option<usize> {
 		self.depth
 	}
+
+	pub fn access_address(&mut self, address: H160) {
+		if let Some(accessed) = &mut self.accessed {
+			accessed.access_address(address)
+		}
+	}
+
+	pub fn access_addresses<I>(&mut self, addresses: I)
+	where
+		I: Iterator<Item = H160>,
+	{
+		if let Some(accessed) = &mut self.accessed {
+			accessed.access_addresses(addresses);
+		}
+	}
+
+	pub fn access_storage(&mut self, address: H160, key: H256) {
+		if let Some(accessed) = &mut self.accessed {
+			accessed.accessed_storage.insert((address, key));
+		}
+	}
+
+	pub fn access_storages<I>(&mut self, storages: I)
+	where
+		I: Iterator<Item = (H160, H256)>,
+	{
+		if let Some(accessed) = &mut self.accessed {
+			accessed.access_storages(storages);
+		}
+	}
 }
 
 #[derive(Debug, Eq, PartialEq, Clone)]
@@ -88,46 +192,37 @@ pub struct PrecompileOutput {
 	pub logs: Vec<Log>,
 }
 
+/// A precompile result.
+pub type PrecompileResult = Result<PrecompileOutput, ExitError>;
+
 /// Precompiles function signature. Expected input arguments are:
-///  * Address
 ///  * Input
 ///  * Context
-pub type PrecompileFn =
-	fn(H160, &[u8], Option<u64>, &Context) -> Option<Result<PrecompileOutput, ExitError>>;
+///  * Is static
+pub type PrecompileFn = fn(&[u8], Option<u64>, &Context, bool) -> PrecompileResult;
+
+/// A map of address keys to precompile function values.
+pub type Precompile = BTreeMap<H160, PrecompileFn>;
 
 /// Stack-based executor.
 pub struct StackExecutor<'config, S> {
 	config: &'config Config,
-	precompile: PrecompileFn,
+	precompile: Precompile,
 	state: S,
 }
 
-fn no_precompile(
-	_address: H160,
-	_input: &[u8],
-	_target_gas: Option<u64>,
-	_context: &Context,
-) -> Option<Result<PrecompileOutput, ExitError>> {
-	None
-}
-
 impl<'config, S: StackState<'config>> StackExecutor<'config, S> {
-	/// Create a new stack-based executor.
-	pub fn new(state: S, config: &'config Config) -> Self {
-		Self::new_with_precompile(state, config, no_precompile)
-	}
-
 	/// Return a reference of the Config.
 	pub fn config(&self) -> &'config Config {
 		self.config
 	}
 
+	pub fn precompile(&self) -> &Precompile {
+		&self.precompile
+	}
+
 	/// Create a new stack-based executor with given precompiles.
-	pub fn new_with_precompile(
-		state: S,
-		config: &'config Config,
-		precompile: PrecompileFn,
-	) -> Self {
+	pub fn new_with_precompile(state: S, config: &'config Config, precompile: Precompile) -> Self {
 		Self {
 			config,
 			precompile,
@@ -181,18 +276,25 @@ impl<'config, S: StackState<'config>> StackExecutor<'config, S> {
 		value: U256,
 		init_code: Vec<u8>,
 		gas_limit: u64,
+		access_list: Vec<(H160, Vec<H256>)>, // See EIP-2930
 	) -> ExitReason {
-		let transaction_cost = gasometer::create_transaction_cost(&init_code);
-		match self
-			.state
-			.metadata_mut()
-			.gasometer
-			.record_transaction(transaction_cost)
-		{
+		event!(TransactCreate {
+			caller,
+			value,
+			init_code: &init_code,
+			gas_limit,
+			address: self.create_address(CreateScheme::Legacy { caller }),
+		});
+
+		let transaction_cost = gasometer::create_transaction_cost(&init_code, &access_list);
+		let gasometer = &mut self.state.metadata_mut().gasometer;
+		match gasometer.record_transaction(transaction_cost) {
 			Ok(()) => (),
-			Err(e) => return e.into(),
+			Err(e) => return emit_exit!(e.into()),
 		}
 
+		self.initialize_with_access_list(access_list);
+
 		match self.create_inner(
 			caller,
 			CreateScheme::Legacy { caller },
@@ -201,7 +303,7 @@ impl<'config, S: StackState<'config>> StackExecutor<'config, S> {
 			Some(gas_limit),
 			false,
 		) {
-			Capture::Exit((s, _, _)) => s,
+			Capture::Exit((s, _, _)) => emit_exit!(s),
 			Capture::Trap(_) => unreachable!(),
 		}
 	}
@@ -214,19 +316,32 @@ impl<'config, S: StackState<'config>> StackExecutor<'config, S> {
 		init_code: Vec<u8>,
 		salt: H256,
 		gas_limit: u64,
+		access_list: Vec<(H160, Vec<H256>)>, // See EIP-2930
 	) -> ExitReason {
-		let transaction_cost = gasometer::create_transaction_cost(&init_code);
-		match self
-			.state
-			.metadata_mut()
-			.gasometer
-			.record_transaction(transaction_cost)
-		{
+		event!(TransactCreate2 {
+			caller,
+			value,
+			init_code: &init_code,
+			salt,
+			gas_limit,
+			address: self.create_address(CreateScheme::Create2 {
+				caller,
+				code_hash: H256::from_slice(Keccak256::digest(&init_code).as_slice()),
+				salt,
+			}),
+		});
+
+		let transaction_cost = gasometer::create_transaction_cost(&init_code, &access_list);
+		let gasometer = &mut self.state.metadata_mut().gasometer;
+		match gasometer.record_transaction(transaction_cost) {
 			Ok(()) => (),
-			Err(e) => return e.into(),
+			Err(e) => return emit_exit!(e.into()),
 		}
+
 		let code_hash = H256::from_slice(Keccak256::digest(&init_code).as_slice());
 
+		self.initialize_with_access_list(access_list);
+
 		match self.create_inner(
 			caller,
 			CreateScheme::Create2 {
@@ -239,12 +354,17 @@ impl<'config, S: StackState<'config>> StackExecutor<'config, S> {
 			Some(gas_limit),
 			false,
 		) {
-			Capture::Exit((s, _, _)) => s,
+			Capture::Exit((s, _, _)) => emit_exit!(s),
 			Capture::Trap(_) => unreachable!(),
 		}
 	}
 
-	/// Execute a `CALL` transaction.
+	/// Execute a `CALL` transaction with a given caller, address, value and
+	/// gas limit and data.
+	///
+	/// Takes in an additional `access_list` parameter for EIP-2930 which was
+	/// introduced in the Ethereum Berlin hard fork. If you do not wish to use
+	/// this functionality, just pass in an empty vector.
 	pub fn transact_call(
 		&mut self,
 		caller: H160,
@@ -252,16 +372,35 @@ impl<'config, S: StackState<'config>> StackExecutor<'config, S> {
 		value: U256,
 		data: Vec<u8>,
 		gas_limit: u64,
+		access_list: Vec<(H160, Vec<H256>)>,
 	) -> (ExitReason, Vec<u8>) {
-		let transaction_cost = gasometer::call_transaction_cost(&data);
-		match self
-			.state
-			.metadata_mut()
-			.gasometer
-			.record_transaction(transaction_cost)
-		{
+		event!(TransactCall {
+			caller,
+			address,
+			value,
+			data: &data,
+			gas_limit,
+		});
+
+		let transaction_cost = gasometer::call_transaction_cost(&data, &access_list);
+		let gasometer = &mut self.state.metadata_mut().gasometer;
+		match gasometer.record_transaction(transaction_cost) {
 			Ok(()) => (),
-			Err(e) => return (e.into(), Vec::new()),
+			Err(e) => return emit_exit!(e.into(), Vec::new()),
+		}
+
+		// Initialize initial addresses for EIP-2929
+		if self.config.increase_state_access_gas {
+			let addresses = self
+				.precompile
+				.clone()
+				.into_iter()
+				.map(|(k, _)| k)
+				.chain(core::iter::once(caller))
+				.chain(core::iter::once(address));
+			self.state.metadata_mut().access_addresses(addresses);
+
+			self.initialize_with_access_list(access_list);
 		}
 
 		self.state.inc_nonce(caller);
@@ -286,7 +425,7 @@ impl<'config, S: StackState<'config>> StackExecutor<'config, S> {
 			false,
 			context,
 		) {
-			Capture::Exit((s, v)) => (s, v),
+			Capture::Exit((s, v)) => emit_exit!(s, v),
 			Capture::Trap(_) => unreachable!(),
 		}
 	}
@@ -337,6 +476,16 @@ impl<'config, S: StackState<'config>> StackExecutor<'config, S> {
 		}
 	}
 
+	pub fn initialize_with_access_list(&mut self, access_list: Vec<(H160, Vec<H256>)>) {
+		let addresses = access_list.iter().map(|a| a.0);
+		self.state.metadata_mut().access_addresses(addresses);
+
+		let storage_keys = access_list
+			.into_iter()
+			.flat_map(|(address, keys)| keys.into_iter().map(move |key| (address, key)));
+		self.state.metadata_mut().access_storages(storage_keys);
+	}
+
 	fn create_inner(
 		&mut self,
 		caller: H160,
@@ -361,6 +510,19 @@ impl<'config, S: StackState<'config>> StackExecutor<'config, S> {
 
 		let address = self.create_address(scheme);
 
+		self.state.metadata_mut().access_address(caller);
+		self.state.metadata_mut().access_address(address);
+
+		let addresses: Vec<H160> = self
+			.precompile
+			.clone()
+			.into_iter()
+			.map(|(k, _)| k)
+			.collect();
+		self.state
+			.metadata_mut()
+			.access_addresses(addresses.iter().copied());
+
 		event!(Create {
 			caller,
 			address,
@@ -583,8 +745,8 @@ impl<'config, S: StackState<'config>> StackExecutor<'config, S> {
 			}
 		}
 
-		if let Some(ret) = (self.precompile)(code_address, &input, Some(gas_limit), &context) {
-			match ret {
+		if let Some(precompile) = self.precompile.get(&code_address) {
+			return match (*precompile)(&input, Some(gas_limit), &context, is_static) {
 				Ok(PrecompileOutput {
 					exit_status,
 					output,
@@ -607,13 +769,13 @@ impl<'config, S: StackState<'config>> StackExecutor<'config, S> {
 
 					let _ = self.state.metadata_mut().gasometer.record_cost(cost);
 					let _ = self.exit_substate(StackExitKind::Succeeded);
-					return Capture::Exit((ExitReason::Succeed(exit_status), output));
+					Capture::Exit((ExitReason::Succeed(exit_status), output))
 				}
 				Err(e) => {
 					let _ = self.exit_substate(StackExitKind::Failed);
-					return Capture::Exit((ExitReason::Error(e), Vec::new()));
+					Capture::Exit((ExitReason::Error(e), Vec::new()))
 				}
-			}
+			};
 		}
 
 		let mut runtime = Runtime::new(Rc::new(code), Rc::new(input), context, self.config);
@@ -687,6 +849,13 @@ impl<'config, S: StackState<'config>> Handler for StackExecutor<'config, S> {
 		}
 	}
 
+	fn is_cold(&self, address: H160, maybe_index: Option<H256>) -> bool {
+		match maybe_index {
+			None => self.state.is_cold(address),
+			Some(index) => self.state.is_storage_cold(address, index),
+		}
+	}
+
 	fn gas_left(&self) -> U256 {
 		U256::from(self.state.metadata().gasometer.gas())
 	}
@@ -753,6 +922,7 @@ impl<'config, S: StackState<'config>> Handler for StackExecutor<'config, S> {
 		Ok(())
 	}
 
+	#[cfg(not(feature = "tracing"))]
 	fn create(
 		&mut self,
 		caller: H160,
@@ -764,6 +934,25 @@ impl<'config, S: StackState<'config>> Handler for StackExecutor<'config, S> {
 		self.create_inner(caller, scheme, value, init_code, target_gas, true)
 	}
 
+	#[cfg(feature = "tracing")]
+	fn create(
+		&mut self,
+		caller: H160,
+		scheme: CreateScheme,
+		value: U256,
+		init_code: Vec<u8>,
+		target_gas: Option<u64>,
+	) -> Capture<(ExitReason, Option<H160>, Vec<u8>), Self::CreateInterrupt> {
+		let capture = self.create_inner(caller, scheme, value, init_code, target_gas, true);
+
+		if let Capture::Exit((ref reason, _, ref return_value)) = capture {
+			emit_exit!(reason, return_value);
+		}
+
+		capture
+	}
+
+	#[cfg(not(feature = "tracing"))]
 	fn call(
 		&mut self,
 		code_address: H160,
@@ -785,6 +974,34 @@ impl<'config, S: StackState<'config>> Handler for StackExecutor<'config, S> {
 		)
 	}
 
+	#[cfg(feature = "tracing")]
+	fn call(
+		&mut self,
+		code_address: H160,
+		transfer: Option<Transfer>,
+		input: Vec<u8>,
+		target_gas: Option<u64>,
+		is_static: bool,
+		context: Context,
+	) -> Capture<(ExitReason, Vec<u8>), Self::CallInterrupt> {
+		let capture = self.call_inner(
+			code_address,
+			transfer,
+			input,
+			target_gas,
+			is_static,
+			true,
+			true,
+			context,
+		);
+
+		if let Capture::Exit((ref reason, ref return_value)) = capture {
+			emit_exit!(reason, return_value);
+		}
+
+		capture
+	}
+
 	#[inline]
 	fn pre_validate(
 		&mut self,
@@ -798,7 +1015,7 @@ impl<'config, S: StackState<'config>> Handler for StackExecutor<'config, S> {
 			self.state.metadata_mut().gasometer.record_cost(cost)?;
 		} else {
 			let is_static = self.state.metadata().is_static;
-			let (gas_cost, memory_cost) = gasometer::dynamic_opcode_cost(
+			let (gas_cost, target, memory_cost) = gasometer::dynamic_opcode_cost(
 				context.address,
 				opcode,
 				stack,
@@ -810,6 +1027,15 @@ impl<'config, S: StackState<'config>> Handler for StackExecutor<'config, S> {
 			let gasometer = &mut self.state.metadata_mut().gasometer;
 
 			gasometer.record_dynamic_cost(gas_cost, memory_cost)?;
+			match target {
+				StorageTarget::Address(address) => {
+					self.state.metadata_mut().access_address(address)
+				}
+				StorageTarget::Slot(address, key) => {
+					self.state.metadata_mut().access_storage(address, key)
+				}
+				StorageTarget::None => (),
+			}
 		}
 
 		Ok(())
