@@ -1,5 +1,5 @@
 use crate::backend::{Apply, Backend, Basic, Log};
-use crate::executor::stack::StackSubstateMetadata;
+use crate::executor::stack::{Accessed, StackSubstateMetadata};
 use crate::{ExitError, Transfer};
 use alloc::{
 	boxed::Box,
@@ -10,12 +10,13 @@ use core::mem;
 use primitive_types::{H160, H256, U256};
 
 #[derive(Clone, Debug)]
-struct MemoryStackAccount {
+pub struct MemoryStackAccount {
 	pub basic: Basic,
 	pub code: Option<Vec<u8>>,
 	pub reset: bool,
 }
 
+#[derive(Clone, Debug)]
 pub struct MemoryStackSubstate<'config> {
 	metadata: StackSubstateMetadata<'config>,
 	parent: Option<Box<MemoryStackSubstate<'config>>>,
@@ -37,6 +38,14 @@ impl<'config> MemoryStackSubstate<'config> {
 		}
 	}
 
+	pub fn logs(&self) -> &[Log] {
+		&self.logs
+	}
+
+	pub fn logs_mut(&mut self) -> &mut Vec<Log> {
+		&mut self.logs
+	}
+
 	pub fn metadata(&self) -> &StackSubstateMetadata<'config> {
 		&self.metadata
 	}
@@ -165,7 +174,7 @@ impl<'config> MemoryStackSubstate<'config> {
 		Ok(())
 	}
 
-	fn known_account(&self, address: H160) -> Option<&MemoryStackAccount> {
+	pub fn known_account(&self, address: H160) -> Option<&MemoryStackAccount> {
 		if let Some(account) = self.accounts.get(&address) {
 			Some(account)
 		} else if let Some(parent) = self.parent.as_ref() {
@@ -237,6 +246,26 @@ impl<'config> MemoryStackSubstate<'config> {
 		None
 	}
 
+	pub fn is_cold(&self, address: H160) -> bool {
+		self.recursive_is_cold(&|a| a.accessed_addresses.contains(&address))
+	}
+
+	pub fn is_storage_cold(&self, address: H160, key: H256) -> bool {
+		self.recursive_is_cold(&|a: &Accessed| a.accessed_storage.contains(&(address, key)))
+	}
+
+	fn recursive_is_cold<F: Fn(&Accessed) -> bool>(&self, f: &F) -> bool {
+		let local_is_accessed = self.metadata.accessed.as_ref().map(f).unwrap_or(false);
+		if local_is_accessed {
+			false
+		} else {
+			self.parent
+				.as_ref()
+				.map(|p| p.recursive_is_cold(f))
+				.unwrap_or(true)
+		}
+	}
+
 	pub fn deleted(&self, address: H160) -> bool {
 		if self.deletes.contains(&address) {
 			return true;
@@ -364,6 +393,7 @@ impl<'config> MemoryStackSubstate<'config> {
 	}
 }
 
+#[auto_impl::auto_impl(&mut, Box)]
 pub trait StackState<'config>: Backend {
 	fn metadata(&self) -> &StackSubstateMetadata<'config>;
 	fn metadata_mut(&mut self) -> &mut StackSubstateMetadata<'config>;
@@ -375,6 +405,8 @@ pub trait StackState<'config>: Backend {
 
 	fn is_empty(&self, address: H160) -> bool;
 	fn deleted(&self, address: H160) -> bool;
+	fn is_cold(&self, address: H160) -> bool;
+	fn is_storage_cold(&self, address: H160, key: H256) -> bool;
 
 	fn inc_nonce(&mut self, address: H160);
 	fn set_storage(&mut self, address: H160, key: H256, value: H256);
@@ -387,6 +419,7 @@ pub trait StackState<'config>: Backend {
 	fn touch(&mut self, address: H160);
 }
 
+#[derive(Clone, Debug)]
 pub struct MemoryStackState<'backend, 'config, B> {
 	backend: &'backend B,
 	substate: MemoryStackSubstate<'config>,
@@ -491,6 +524,14 @@ impl<'backend, 'config, B: Backend> StackState<'config> for MemoryStackState<'ba
 		self.substate.deleted(address)
 	}
 
+	fn is_cold(&self, address: H160) -> bool {
+		self.substate.is_cold(address)
+	}
+
+	fn is_storage_cold(&self, address: H160, key: H256) -> bool {
+		self.substate.is_storage_cold(address, key)
+	}
+
 	fn inc_nonce(&mut self, address: H160) {
 		self.substate.inc_nonce(address, self.backend);
 	}
